// Michael J. Wilson
// Music 256a Fall 2010
// Homework #4

#include <cmath>
#include "ClickGenerator.h"
#include "Utility.h"

ClickGenerator::~ClickGenerator()
{
    pthread_mutex_unlock(&m_mutex);
    pthread_mutex_destroy(&m_mutex);
}

ClickGenerator::ClickGenerator()
{
    m_scheduled_clicks.clear();
    m_remaining_silence = 0;
    pthread_mutex_init(&m_mutex, NULL);
}

void ClickGenerator::lock()
{
    pthread_mutex_lock(&m_mutex);
}

void ClickGenerator::unlock()
{
    pthread_mutex_unlock(&m_mutex);
}

void ClickGenerator::schedule_clicks(unsigned int num_clicks)
{
    // Assumes we have already locked the mutex
    m_scheduled_clicks.push_back(num_clicks);
}

float ClickGenerator::get_next_sample()
{
    /// Delay between clicks, in seconds
    const float click_delay = .065;
   
    /// Delay between click clusters, in seconds
    const float click_cluster_delay = 0.3;

    /// Length of one click, in seconds (should be less than 80 microseconds)
    const float click_length = .0007;

    // Assumes we have already locked the mutex
    float result = 0.0;
    if (m_remaining_silence > 0)
    {
        m_remaining_silence--;
    }
    else if (m_remaining_clicks > 0)
    {
        result = unit_click();
        m_remaining_clicks--;
        if (m_remaining_clicks == 0)
        {
            m_remaining_silence = Utility::sample_rate * click_delay;
        }
    }
    else if (!m_scheduled_clicks.empty())
    {
        if (m_scheduled_clicks.front() > 0)
        {
            m_remaining_clicks = Utility::sample_rate * click_length;
            m_scheduled_clicks.front()--;
        }
        else
        {
            m_scheduled_clicks.pop_front();
            m_remaining_silence = Utility::sample_rate * click_cluster_delay;
        }
    }

    m_current_sample++;

    return(result);
}

float ClickGenerator::unit_click()
{
    // Dolphin click should be bimodal: a low-frequency peak occurs
    // between 30 and 50 kHz, and a high-frequency peak between 80 and
    // 120 kHz.  This will alias of course, but that's ok.  It's so
    // short we expect a wideband spectrum anyway.
    float result = 0.0;
    const float frequency1 = 45000;
    const float frequency2 = 89000;

    result += std::sin(2 * Utility::PI * frequency1 * ((float)m_current_sample) / (double)Utility::sample_rate);
    result += std::sin(2 * Utility::PI * frequency2 * ((float)m_current_sample) / (double)Utility::sample_rate);
    result /= 2;

    return(result);
}



