// Michael J. Wilson
// Music 256a Fall 2010
// Homework #4

#ifndef __UTILITY_H__
#define __UTILITY_H__

#include <string>

class RtAudio;

/// Singleton utility class, contains data that should be available to
/// the entire program
class Utility
{
public:
    /// Destructor
    ~Utility();

    /// Get the utility instance.  Lazily constructed.
    static Utility* get_utility();

    /// Ratio of circle's circumferance to its diameter
    static const float PI;

    /// Very small value for floating point comparisons
    static const float EPSILON;

    /// Sample rate
    static const unsigned int MAXIMUM_SAMPLE_RATE;

    /// Sample rate
    static unsigned int sample_rate;

    /// Number of output channels
    static const unsigned int OUTPUT_CHANNELS;

    /// Name of the program, what the user would type at the command line
    static const std::string PROGRAM_NAME;

    /// A pointer to the DAC
    RtAudio* dac;

    /// Get a random number in the specified range.  No assurances on
    /// its distribution.
    float rand(float min, float max);

    // Get the frequency of the specified MIDI note
    static float midi_to_frequency(int note);

    /// Get the closest MIDI note to the specified frequency
    static int frequency_to_midi(float frequency);

    /// Get the pitch name for the specified MIDI note
    static std::string midi_to_string(int note);

private:
    /// Private constructor to follow singleton pattern
    Utility();
};

#endif
