// Michael J. Wilson
// Music 256a Fall 2010
// Homework #4

#include <algorithm>
#include <cctype>
#include <cstdlib>
#include <pthread.h>
#include <string>
#include <cstring>
#include <sstream>

#include "osc/OscOutboundPacketStream.h"
#include "osc/OscPacketListener.h"
#include "osc/OscReceivedElements.h"
#include "ip/UdpSocket.h"
#include "RtAudio.h"

#include "ClickGenerator.h"
#include "Utility.h"

/// Click buffers
ClickGenerator local_clicks;
ClickGenerator remote_clicks;

/// Hostname to connect to
std::string hostname("localhosgatt");

/// Fixed port to connect to
const int port = 7000;

//////////////////////////////////////////////////////////////////////
/// Packet listener class
//////////////////////////////////////////////////////////////////////
class OscReceiveTestPacketListener : public osc::OscPacketListener
{
protected:
    void ProcessMessage(const osc::ReceivedMessage& m, const IpEndpointName& remoteEndpoint )
        {
            try
            {
                osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                if( strcmp( m.AddressPattern(), "/click" ) == 0 )
                {
                    osc::int32 a;
                    args >> a >> osc::EndMessage;
                    remote_clicks.schedule_clicks(a);
                }
            }
            catch(osc::Exception& e)
            {
                std::cout << "error while parsing message: "
                          << m.AddressPattern() << ": " << e.what() << "\n";
            }
        }    
};

//////////////////////////////////////////////////////////////////////
/// Callback for listening to incomming OSC messages
//////////////////////////////////////////////////////////////////////
void *osc_listener(void *threadid)
{
    OscReceiveTestPacketListener listener;
    UdpListeningReceiveSocket s(
        IpEndpointName( IpEndpointName::ANY_ADDRESS, port ),
        &listener );
    s.RunUntilSigInt();
    pthread_exit(NULL);
}

//////////////////////////////////////////////////////////////////////
/// Audio callback
//////////////////////////////////////////////////////////////////////
int audio_callback(void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames, double streamTime, RtAudioStreamStatus status, void *userData)
{
    float* obuffer = (float*)outputBuffer;
    unsigned int i;
    unsigned int j;
    float next_sample;

    if (status)
    {
        std::cout << "Stream underflow detected!" << std::endl;
    }

    for (i = 0; i < nBufferFrames; i++)
    {
        next_sample = (local_clicks.get_next_sample() * 2/3) + (remote_clicks.get_next_sample() * 1/3);
        for (j = 0; j < Utility::OUTPUT_CHANNELS; j++)
        {
            *obuffer++ = next_sample;
        }
    }

    return(0);
}

//////////////////////////////////////////////////////////////////////
/// Send clicks
//////////////////////////////////////////////////////////////////////
void send_clicks(unsigned int num_clicks)
{
    const int IP_MTU_SIZE = 1536;
    IpEndpointName host(hostname.c_str(), port);
    char buffer[IP_MTU_SIZE];
    osc::OutboundPacketStream p( buffer, IP_MTU_SIZE );
    UdpTransmitSocket socket( host );
    p.Clear();
    p << osc::BeginMessage("/click") <<(osc::int32) num_clicks << osc::EndMessage;
    socket.Send( p.Data(), p.Size() );
}

//////////////////////////////////////////////////////////////////////
/// Main input loop
//////////////////////////////////////////////////////////////////////
void command_interpreter()
{
    std::cout << "WELCOME TO DOLPHINS" << std::endl
              << "                            ..                                              " << std::endl
              << "                          :`   ``´:..                                       " << std::endl
              << "                           ´:`:.    ``´:.                                   " << std::endl
              << "                             : `-:.     `´:.                                " << std::endl
              << "                              : ::´    .::::::::...                         " << std::endl
              << "                              : :´`               ``´::.                    " << std::endl
              << "                             :´                        `´::                 " << std::endl
              << "                           :´                              `:.              " << std::endl
              << "                         .´                                  `´:            " << std::endl
              << "                        :                                       ´:          " << std::endl
              << "                      :´                                          ´:        " << std::endl
              << "                    .:                                              :.      " << std::endl
              << "                   :´                                                `:     " << std::endl
              << "                  :´                                                   :    " << std::endl
              << "                 :                                                      :   " << std::endl
              << "                :                 .::::::::         .:                   :  " << std::endl
              << "               :              .:::::::´´´´;        :`                    ;  " << std::endl
              << "              :            ::::´`        :        :``::. .        :.      : " << std::endl
              << "             .´         .::´`           :       .´     `´:::.      `      : " << std::endl
              << "             :        .:´`             ;      .:           `´:   .   :    : " << std::endl
              << "            :        .´              :`  ..::´                `: `-.  `:..: " << std::endl
              << "            ;      .:                 ````                      `:. ´:. ´:  " << std::endl
              << "           :      :´                                              `´;.`:  : " << std::endl
              << "           :     :                                                   `:.:. :" << std::endl
              << "         :´´´   :                                                      `´::;" << std::endl
              << "       :´       :                                                           " << std::endl
              << "      :        ':                                                           " << std::endl
              << "     ´      .   ´.                               ASCII art by Cetus         " << std::endl
              << "    ;   :::`:.   `.                                                         " << std::endl
              << "   :  :´      `´. :                                                         " << std::endl
              << std::endl
              << std::endl
              << std::endl
              << "YOU ARE SWIMMING ALONE." << std::endl
              << "Type 'connect' followed by an IP address to connect to another dolphin." << std::endl
              << "Type 'click' followed by some numbers to send out echolocation clicks." << std::endl
              << "Type 'quit' to quit." << std::endl;

    bool quit = false;
    std::string command;
    std::string token;
    unsigned int clicks;
    while (!quit)
    {
        // Get user input and convert to uppercase
        std::getline(std::cin,command);
        std::stringstream ss(command);
        ss >> token;
        std::transform(token.begin(), token.end(), token.begin(), (int(*)(int)) std::tolower);
        if (token.compare("quit") == 0)
        {
            quit = true;
        }
        else if (token.compare("click") == 0)
        {
            while (!ss.eof() && !ss.fail())
            {
                ss >> token;
                std::stringstream converter(token);
                if (!(converter >> clicks).fail())
                {
                    local_clicks.schedule_clicks(clicks);
                    send_clicks(clicks);
                }
            }
        }
        else if (token.compare("connect") == 0)
        {
            if (!ss.eof() && !ss.fail())
            {
                ss >> token;
                hostname = token;
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////
/// Handle main input
//////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    bool error_occured = false;
    unsigned int buffer_frames = 512;
    pthread_t listener_thread;
    int return_code;

    //////////////////////////////////////////////////
    // Set up audio
    if (Utility::get_utility()->dac->getDeviceCount() < 1)
    {
        std::cout << std::endl << "No audio devices found!" << std::endl;
        error_occured = true;
    }
    if (!error_occured)
    {
        // Get the best sample rate, up to our maximum:
        Utility::sample_rate = 0;
        RtAudio::DeviceInfo info = Utility::get_utility()->dac->getDeviceInfo(Utility::get_utility()->dac->getDefaultInputDevice());
        for (std::vector<unsigned int>::iterator sr_iter = info.sampleRates.begin(); sr_iter != info.sampleRates.end(); ++sr_iter)
        {
            if (*sr_iter <= Utility::MAXIMUM_SAMPLE_RATE && *sr_iter > Utility::sample_rate)
            {
                Utility::sample_rate = *sr_iter;
            }
        }

        RtAudio::StreamParameters output_parameters;
        output_parameters.deviceId = Utility::get_utility()->dac->getDefaultOutputDevice();
        output_parameters.nChannels = Utility::OUTPUT_CHANNELS;
        output_parameters.firstChannel = 0;

        try
        {
            Utility::get_utility()->dac->openStream(
                &output_parameters,
                NULL,
                RTAUDIO_FLOAT32,
                Utility::sample_rate,
                &buffer_frames,
                &audio_callback,
                NULL);
            Utility::get_utility()->dac->startStream();
        }
        catch (RtError& e)
        {
            e.printMessage();
            error_occured = true;
        }
    }

    //////////////////////////////////////////////////
    // Start OSC listener
    if (!error_occured)
    {
        return_code = pthread_create(&listener_thread, NULL, osc_listener, NULL);
        if(return_code)
        {
            std::cout << "Error " << return_code << " when creating listener thread!" << std::endl;
            error_occured = true;
        }
    }

    //////////////////////////////////////////////////
    // Start main loop
    if (!error_occured)
    {
        command_interpreter();
    }

    //////////////////////////////////////////////////
    // Cleanup
    if (Utility::get_utility()->dac->isStreamRunning())
    {
        try
        {
            // Stop the stream
            Utility::get_utility()->dac->stopStream();
        }
        catch (RtError& e)
        {
            e.printMessage();
        }
    }

    if (Utility::get_utility()->dac->isStreamOpen())
    {
        Utility::get_utility()->dac->closeStream();
    }

    return (error_occured ? EXIT_FAILURE : EXIT_SUCCESS);
}
